const backtrace = (current, from)=>{
    const path = [
        current
    ];
    let next;
    while(next = from.get(current)){
        path.push(next);
        current = next;
    }
    return path;
};
const distanceFnMap = new WeakMap();
const getDistance = ([fn, scale = 1], current, neighbor, from)=>{
    let neighborsMap = distanceFnMap.get(fn);
    if (!neighborsMap) neighborsMap = new WeakMap();
    let neighbors = neighborsMap.get(current);
    if (!neighbors) neighbors = new Map();
    let dist = neighbors.get(neighbor);
    if (dist == null) {
        dist = fn(current, neighbor, from) * scale;
        neighbors.set(neighbor, dist);
    }
    return dist;
};
const getAgent = ({ start , goal , hash , heuristic: [heuristic, scale = 1] , neighbors , distance , negativeHeap =false  })=>{
    const add = negativeHeap ? -1 : 1;
    const from = new Map();
    const G = new Map();
    G.set(start, 0);
    const F = new Map();
    F.set(start, heuristic(start, goal) * scale);
    const heap = {
        0: start
    };
    const opened = {
        [hash(start)]: start
    };
    return {
        from,
        get: (node)=>opened[hash(node)],
        peek: ()=>{
            let current;
            for(const key in heap){
                current = heap[key];
                delete heap[key];
                break;
            }
            return current;
        },
        advance: (current)=>{
            const cg = G.get(current) ?? Infinity;
            for (const neighbor of neighbors(current)){
                const key = hash(neighbor);
                if (key in opened) continue;
                opened[key] = neighbor;
                const g = cg + getDistance(distance, current, neighbor, from);
                const cheapest = G.get(neighbor) ?? Infinity;
                if (g < cheapest) {
                    from.set(neighbor, current);
                    G.set(neighbor, g);
                    let f = g + heuristic(neighbor, goal) * scale << 4;
                    while(f in heap)f += add;
                    F.set(neighbor, f);
                    heap[f] = neighbor;
                }
            }
        }
    };
};
export const findShortestPath = ({ start , goal , hash , neighbors , strategies , startedAt =performance.now() , attempt =0 , totalIterations =0  })=>{
    const { accept , heuristic , distance , negativeHeap , maxIterations =200  } = strategies[attempt++];
    const agent = getAgent({
        start,
        goal,
        hash,
        heuristic,
        neighbors,
        distance,
        negativeHeap
    });
    let current;
    let iterations = 0;
    while(++iterations < maxIterations && (current = agent.peek())){
        if (accept(current, goal, agent.from, agent.from)) {
            const result = [
                goal
            ].concat(backtrace(current, agent.from)).reverse();
            result.meta = {
                attempt,
                iterations,
                totalIterations: totalIterations + iterations,
                totalTimeMs: performance.now() - startedAt
            };
            return result;
        }
        agent.advance(current);
    }
    if (attempt < strategies.length) {
        return findShortestPath({
            start,
            goal,
            hash,
            neighbors,
            strategies,
            attempt,
            startedAt,
            totalIterations: totalIterations + iterations
        });
    }
    const result1 = [];
    result1.meta = {
        attempt,
        iterations,
        totalIterations: totalIterations + iterations,
        totalTimeMs: performance.now() - startedAt
    };
    return result1;
};
export const findShortestPathBi = ({ start , goal , hash , strategies , neighbors , startedAt =performance.now() , attempt =0 , totalIterations =0  })=>{
    const { accept , heuristic , distance , negativeHeap , maxIterations =200  } = strategies[attempt++];
    const startAgent = getAgent({
        start,
        goal,
        hash,
        heuristic,
        neighbors,
        distance,
        negativeHeap
    });
    const goalAgent = getAgent({
        start: goal,
        goal: start,
        hash,
        heuristic,
        neighbors,
        distance,
        negativeHeap
    });
    let currentStart = start;
    let currentGoal = goal;
    let lastStart = start;
    let lastGoal = goal;
    let maybeStart;
    let maybeGoal;
    let iterations = 0;
    while(++iterations < maxIterations){
        if (currentStart) {
            currentStart = startAgent.peek();
            if (currentStart) {
                lastStart = currentStart;
            }
        }
        if (currentGoal) {
            currentGoal = goalAgent.peek();
            if (currentGoal) {
                lastGoal = currentGoal;
            }
        }
        maybeStart = startAgent.get(lastGoal);
        maybeGoal = goalAgent.get(lastStart);
        if (maybeStart) lastStart = maybeStart;
        if (maybeGoal) lastGoal = maybeGoal;
        if (maybeStart || maybeGoal || accept(lastStart, lastGoal, startAgent.from, goalAgent.from)) {
            let startPath = backtrace(lastStart, startAgent.from).reverse();
            let goalPath = backtrace(lastGoal, goalAgent.from);
            if (goalPath.length > 1) goalPath = goalPath.slice(1);
            else startPath = startPath.slice(0, -1);
            const result = startPath.concat(goalPath);
            result.meta = {
                attempt,
                iterations,
                totalIterations: totalIterations + iterations,
                totalTimeMs: performance.now() - startedAt
            };
            return result;
        }
        if (currentStart) {
            startAgent.advance(currentStart);
        }
        if (currentGoal) {
            goalAgent.advance(currentGoal);
        }
    }
    if (attempt < strategies.length) {
        return findShortestPathBi({
            start,
            goal,
            hash,
            neighbors,
            strategies,
            attempt,
            startedAt,
            totalIterations: totalIterations + iterations
        });
    }
    const result1 = [];
    result1.meta = {
        attempt,
        iterations,
        totalIterations: totalIterations + iterations,
        totalTimeMs: performance.now() - startedAt
    };
    return result1;
};

//# sourceMappingURL=find-shortest-path.js.map