declare type Accept<T> = (currentStart: T, currentGoal: T, startFrom: Map<T, T>, goalFrom: Map<T, T>) => boolean;
declare type Hash<T> = (node: T) => string;
declare type Heuristic<T> = (node: T, goal: T) => number;
declare type Neighbors<T> = (node: T) => T[];
declare type Distance<T> = (current: T, neighbor: T, from: Map<T, T>) => number;
declare type Scaled<T> = [T, number?];
declare type ShortestPathMeta = {
    attempt: number;
    iterations: number;
    totalIterations: number;
    totalTimeMs: number;
};
export interface ShortestPath<T> extends Array<T> {
    meta: ShortestPathMeta;
}
interface AgentStrategy<T> {
    distance: Scaled<Distance<T>>;
    heuristic: Scaled<Heuristic<T>>;
    maxIterations?: number;
    negativeHeap?: boolean;
}
export interface Strategy<T> extends AgentStrategy<T> {
    accept: Accept<T>;
}
interface FindShortestPathAgentOptions<T> {
    start: T;
    goal: T;
    hash: Hash<T>;
    neighbors: Neighbors<T>;
}
interface FindShortestPathOptions<T> extends FindShortestPathAgentOptions<T> {
    strategies: Strategy<T>[];
    attempt?: number;
    totalIterations?: number;
    startedAt?: number;
}
export declare const findShortestPath: <T>({ start, goal, hash, neighbors, strategies, startedAt, attempt, totalIterations }: FindShortestPathOptions<T>) => ShortestPath<T>;
export declare const findShortestPathBi: <T>({ start, goal, hash, strategies, neighbors, startedAt, attempt, totalIterations, }: FindShortestPathOptions<T>) => ShortestPath<T>;
export {};
